@@ -1,7 +1,7 @@
 /*
  *  DHCP library with GLib integration
  *
- *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -155,14 +155,16 @@ uint8_t *dhcpv6_get_option(struct dhcpv6_packet *packet, uint16_t pkt_len,
 	rem = pkt_len - 1 - 3;
 
 	if (rem <= 0)
-		/* Bad packet */
-		return NULL;
+		goto bad_packet;
 
 	while (1) {
 		opt_code = optionptr[0] << 8 | optionptr[1];
 		opt_len = len = optionptr[2] << 8 | optionptr[3];
 		len += 2 + 2; /* skip code and len */
 
+		if (len < 4)
+			goto bad_packet;
+
 		rem -= len;
 		if (rem < 0)
 			break;
@@ -170,7 +172,10 @@ uint8_t *dhcpv6_get_option(struct dhcpv6_packet *packet, uint16_t pkt_len,
 		if (opt_code == code) {
 			if (option_len != NULL)
 				*option_len = opt_len;
-			found = optionptr + 2 + 2;
+			if (rem == 0)
+				found = NULL;
+			else
+				found = optionptr + 2 + 2;
 			count++;
 		}
 
@@ -184,6 +189,13 @@ uint8_t *dhcpv6_get_option(struct dhcpv6_packet *packet, uint16_t pkt_len,
 		*option_count = count;
 
 	return found;
+
+bad_packet:
+	if (option_len != NULL)
+		*option_len = 0;
+	if (option_count != NULL)
+		*option_count = 0;
+	return NULL;
 }
 
 uint8_t *dhcpv6_get_sub_option(unsigned char *option, uint16_t max_len,
@@ -254,25 +266,70 @@ void dhcpv6_add_binary_option(struct dhcpv6_packet *packet, uint16_t max_len,
 	*pkt_len += len;
 }
 
-void dhcp_add_simple_option(struct dhcp_packet *packet, uint8_t code,
-							uint32_t data)
+static GDHCPOptionType check_option(uint8_t code, uint8_t data_len)
 {
-	uint8_t option[6], len;
 	GDHCPOptionType type = dhcp_get_code_type(code);
+	uint8_t len;
 
 	if (type == OPTION_UNKNOWN)
+		return type;
+
+	len = dhcp_option_lengths[type & OPTION_TYPE_MASK];
+	if (len != data_len) {
+		printf("Invalid option len %d (expecting %d) for code 0x%x\n",
+			data_len, len, code);
+		return OPTION_UNKNOWN;
+	}
+
+	return type;
+}
+
+void dhcp_add_option_uint32(struct dhcp_packet *packet, uint8_t code,
+							uint32_t data)
+{
+	uint8_t option[6];
+
+	if (check_option(code, sizeof(data)) == OPTION_UNKNOWN)
 		return;
 
 	option[OPT_CODE] = code;
+	option[OPT_LEN] = sizeof(data);
+	put_be32(data, option + OPT_DATA);
 
-	len = dhcp_option_lengths[type & OPTION_TYPE_MASK];
-	option[OPT_LEN] = len;
+	dhcp_add_binary_option(packet, option);
 
-#if __BYTE_ORDER == __BIG_ENDIAN
-	data <<= 8 * (4 - len);
-#endif
+	return;
+}
+
+void dhcp_add_option_uint16(struct dhcp_packet *packet, uint8_t code,
+							uint16_t data)
+{
+	uint8_t option[6];
+
+	if (check_option(code, sizeof(data)) == OPTION_UNKNOWN)
+		return;
+
+	option[OPT_CODE] = code;
+	option[OPT_LEN] = sizeof(data);
+	put_be16(data, option + OPT_DATA);
+
+	dhcp_add_binary_option(packet, option);
+
+	return;
+}
+
+void dhcp_add_option_uint8(struct dhcp_packet *packet, uint8_t code,
+							uint8_t data)
+{
+	uint8_t option[6];
+
+	if (check_option(code, sizeof(data)) == OPTION_UNKNOWN)
+		return;
+
+	option[OPT_CODE] = code;
+	option[OPT_LEN] = sizeof(data);
+	option[OPT_DATA] = data;
 
-	dhcp_put_unaligned(data, (uint32_t *) &option[OPT_DATA]);
 	dhcp_add_binary_option(packet, option);
 
 	return;
@@ -296,7 +353,7 @@ void dhcp_init_header(struct dhcp_packet *packet, char type)
 	packet->cookie = htonl(DHCP_MAGIC);
 	packet->options[0] = DHCP_END;
 
-	dhcp_add_simple_option(packet, DHCP_MESSAGE_TYPE, type);
+	dhcp_add_option_uint8(packet, DHCP_MESSAGE_TYPE, type);
 }
 
 void dhcpv6_init_header(struct dhcpv6_packet *packet, uint8_t type)
@@ -563,7 +620,7 @@ int dhcp_send_kernel_packet(struct dhcp_packet *dhcp_pkt,
 	memset(&client, 0, sizeof(client));
 	client.sin_family = AF_INET;
 	client.sin_port = htons(source_port);
-	client.sin_addr.s_addr = source_ip;
+	client.sin_addr.s_addr = htonl(source_ip);
 	if (bind(fd, (struct sockaddr *) &client, sizeof(client)) < 0) {
 		close(fd);
 		return -errno;
@@ -572,7 +629,7 @@ int dhcp_send_kernel_packet(struct dhcp_packet *dhcp_pkt,
 	memset(&client, 0, sizeof(client));
 	client.sin_family = AF_INET;
 	client.sin_port = htons(dest_port);
-	client.sin_addr.s_addr = dest_ip;
+	client.sin_addr.s_addr = htonl(dest_ip);
 	if (connect(fd, (struct sockaddr *) &client, sizeof(client)) < 0) {
 		close(fd);
 		return -errno;
