@@ -36,7 +36,7 @@ static DBusConnection *connection;
 static GHashTable *session_hash;
 static connman_bool_t sessionmode;
 static struct connman_session *ecall_session;
-static struct connman_session_config *session_config;
+static GSList *policy_list;
 
 enum connman_session_trigger {
 	CONNMAN_SESSION_TRIGGER_UNKNOWN		= 0,
@@ -59,21 +59,6 @@ enum connman_session_state {
 	CONNMAN_SESSION_STATE_ONLINE         = 2,
 };
 
-enum connman_session_type {
-	CONNMAN_SESSION_TYPE_ANY      = 0,
-	CONNMAN_SESSION_TYPE_LOCAL    = 1,
-	CONNMAN_SESSION_TYPE_INTERNET = 2,
-};
-
-enum connman_session_roaming_policy {
-	CONNMAN_SESSION_ROAMING_POLICY_UNKNOWN		= 0,
-	CONNMAN_SESSION_ROAMING_POLICY_DEFAULT		= 1,
-	CONNMAN_SESSION_ROAMING_POLICY_ALWAYS		= 2,
-	CONNMAN_SESSION_ROAMING_POLICY_FORBIDDEN	= 3,
-	CONNMAN_SESSION_ROAMING_POLICY_NATIONAL		= 4,
-	CONNMAN_SESSION_ROAMING_POLICY_INTERNATIONAL	= 5,
-};
-
 struct service_entry {
 	/* track why this service was selected */
 	enum connman_session_reason reason;
@@ -86,13 +71,8 @@ struct service_entry {
 };
 
 struct session_info {
+	struct connman_session_config config;
 	enum connman_session_state state;
-	enum connman_session_type type;
-	connman_bool_t priority;
-	GSList *allowed_bearers;
-	connman_bool_t avoid_handover;
-	enum connman_session_roaming_policy roaming_policy;
-
 	struct service_entry *entry;
 	enum connman_session_reason reason;
 };
@@ -103,9 +83,13 @@ struct connman_session {
 	char *notify_path;
 	guint notify_watch;
 
+	struct connman_session_policy *policy;
+
 	connman_bool_t append_all;
 	struct session_info *info;
 	struct session_info *info_last;
+	struct connman_session_config *policy_config;
+	GSList *user_allowed_bearers;
 
 	connman_bool_t ecall;
 
@@ -113,12 +97,6 @@ struct connman_session {
 	GHashTable *service_hash;
 };
 
-struct bearer_info {
-	char *name;
-	connman_bool_t match_all;
-	enum connman_service_type service_type;
-};
-
 static const char *trigger2string(enum connman_session_trigger trigger)
 {
 	switch (trigger) {
@@ -170,8 +148,10 @@ static const char *state2string(enum connman_session_state state)
 static const char *type2string(enum connman_session_type type)
 {
 	switch (type) {
-	case CONNMAN_SESSION_TYPE_ANY:
+	case CONNMAN_SESSION_TYPE_UNKNOWN:
 		return "";
+	case CONNMAN_SESSION_TYPE_ANY:
+		return "any";
 	case CONNMAN_SESSION_TYPE_LOCAL:
 		return "local";
 	case CONNMAN_SESSION_TYPE_INTERNET:
@@ -181,17 +161,7 @@ static const char *type2string(enum connman_session_type type)
 	return NULL;
 }
 
-static enum connman_session_type string2type(const char *type)
-{
-	if (g_strcmp0(type, "local") == 0)
-		return CONNMAN_SESSION_TYPE_LOCAL;
-	else if (g_strcmp0(type, "internet") == 0)
-		return CONNMAN_SESSION_TYPE_INTERNET;
-
-	return CONNMAN_SESSION_TYPE_ANY;
-}
-
-static enum connman_session_roaming_policy string2roamingpolicy(const char *policy)
+enum connman_session_roaming_policy connman_session_parse_roaming_policy(const char *policy)
 {
 	if (g_strcmp0(policy, "default") == 0)
 		return CONNMAN_SESSION_ROAMING_POLICY_DEFAULT;
@@ -207,25 +177,36 @@ static enum connman_session_roaming_policy string2roamingpolicy(const char *poli
 		return CONNMAN_SESSION_ROAMING_POLICY_UNKNOWN;
 }
 
-static enum connman_service_type bearer2service(const char *bearer)
+enum connman_session_type connman_session_parse_connection_type(const char *type)
 {
-	if (bearer == NULL)
-		return CONNMAN_SERVICE_TYPE_UNKNOWN;
+	if (g_strcmp0(type, "any") == 0)
+		return CONNMAN_SESSION_TYPE_ANY;
+	if (g_strcmp0(type, "local") == 0)
+		return CONNMAN_SESSION_TYPE_LOCAL;
+	else if (g_strcmp0(type, "internet") == 0)
+		return CONNMAN_SESSION_TYPE_INTERNET;
 
+	return CONNMAN_SESSION_TYPE_UNKNOWN;
+}
+
+static int bearer2service(const char *bearer, enum connman_service_type *type)
+{
 	if (g_strcmp0(bearer, "ethernet") == 0)
-		return CONNMAN_SERVICE_TYPE_ETHERNET;
+		*type = CONNMAN_SERVICE_TYPE_ETHERNET;
 	else if (g_strcmp0(bearer, "wifi") == 0)
-		return CONNMAN_SERVICE_TYPE_WIFI;
-	else if (g_strcmp0(bearer, "wimax") == 0)
-		return CONNMAN_SERVICE_TYPE_WIMAX;
+		*type = CONNMAN_SERVICE_TYPE_WIFI;
 	else if (g_strcmp0(bearer, "bluetooth") == 0)
-		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
+		*type = CONNMAN_SERVICE_TYPE_BLUETOOTH;
 	else if (g_strcmp0(bearer, "cellular") == 0)
-		return CONNMAN_SERVICE_TYPE_CELLULAR;
+		*type = CONNMAN_SERVICE_TYPE_CELLULAR;
 	else if (g_strcmp0(bearer, "vpn") == 0)
-		return CONNMAN_SERVICE_TYPE_VPN;
+		*type = CONNMAN_SERVICE_TYPE_VPN;
+	else if (g_strcmp0(bearer, "*") == 0)
+		*type = CONNMAN_SERVICE_TYPE_UNKNOWN;
 	else
-		return CONNMAN_SERVICE_TYPE_UNKNOWN;
+		return -EINVAL;
+
+	return 0;
 }
 
 static char *service2bearer(enum connman_service_type type)
@@ -235,8 +216,6 @@ static char *service2bearer(enum connman_service_type type)
 		return "ethernet";
 	case CONNMAN_SERVICE_TYPE_WIFI:
 		return "wifi";
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-		return "wimax";
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 		return "bluetooth";
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
@@ -253,92 +232,337 @@ static char *service2bearer(enum connman_service_type type)
 	return "";
 }
 
-static int config_get_bool(const char *id, const char *key, connman_bool_t *val)
+static void destroy_policy_config(struct connman_session *session)
 {
-	if (session_config == NULL) {
-		*val = FALSE;
-		return -EINVAL;
+	if (session->policy == NULL) {
+		g_free(session->policy_config);
+		return;
 	}
 
-	return (*session_config->get_bool)(id, key, val);
+	(*session->policy->destroy)(session);
 }
 
-static int config_get_string(const char *id, const char *key, char **val)
+static void free_session(struct connman_session *session)
 {
-	if (session_config == NULL) {
-		*val = NULL;
-		return -EINVAL;
+	if (session == NULL)
+		return;
+
+	if (session->notify_watch > 0)
+		g_dbus_remove_watch(connection, session->notify_watch);
+
+	destroy_policy_config(session);
+	g_slist_free(session->info->config.allowed_bearers);
+	g_free(session->owner);
+	g_free(session->session_path);
+	g_free(session->notify_path);
+	g_free(session->info);
+	g_free(session->info_last);
+
+	g_free(session);
+}
+
+static void cleanup_session(gpointer user_data)
+{
+	struct connman_session *session = user_data;
+	struct session_info *info = session->info;
+
+	DBG("remove %s", session->session_path);
+
+	g_slist_free(session->user_allowed_bearers);
+	g_hash_table_destroy(session->service_hash);
+	g_sequence_free(session->service_list);
+
+	if (info->entry != NULL &&
+			info->entry->reason == CONNMAN_SESSION_REASON_CONNECT) {
+		__connman_service_disconnect(info->entry->service);
 	}
 
-	return (*session_config->get_string)(id, key, val);
+	free_session(session);
 }
 
-static void cleanup_bearer_info(gpointer data, gpointer user_data)
+static int assign_policy_plugin(struct connman_session *session)
 {
-	struct bearer_info *info = data;
+	if (session->policy != NULL)
+		return -EALREADY;
 
-	g_free(info->name);
-	g_free(info);
+	if (policy_list == NULL)
+		return 0;
+
+	session->policy = policy_list->data;
+
+	return 0;
 }
 
-static GSList *session_parse_allowed_bearers(DBusMessageIter *iter)
+struct user_config {
+	DBusMessage *pending;
+
+	enum connman_session_type type;
+	GSList *allowed_bearers;
+};
+
+static void cleanup_user_config(struct user_config *user_config)
+{
+	if (user_config == NULL)
+		return;
+
+	if (user_config->pending != NULL)
+		dbus_message_unref(user_config->pending);
+
+	g_slist_free(user_config->allowed_bearers);
+	g_free(user_config);
+}
+
+static int create_policy_config(struct connman_session *session,
+				connman_session_config_cb callback,
+				struct user_config *user_config)
+{
+	struct connman_session_config *config;
+
+	if (session->policy == NULL) {
+		config = connman_session_create_default_config();
+		if (config == NULL) {
+			free_session(session);
+			cleanup_user_config(user_config);
+			return -ENOMEM;
+		}
+
+		return callback(session, config, user_config, 0);
+	}
+
+	return (*session->policy->create)(session, callback, user_config);
+}
+
+static void probe_policy(struct connman_session_policy *policy)
+{
+
+	GHashTableIter iter;
+	gpointer key, value;
+	struct connman_session *session;
+
+	DBG("policy %p name %s", policy, policy->name);
+
+	g_hash_table_iter_init(&iter, session_hash);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		session = value;
+
+		if (session->policy != NULL)
+			continue;
+
+		assign_policy_plugin(session);
+	}
+}
+
+static void remove_policy(struct connman_session_policy *policy)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+	struct connman_session *session;
+
+	DBG("policy %p name %s", policy, policy->name);
+
+	g_hash_table_iter_init(&iter, session_hash);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		session = value;
+
+		if (session->policy != policy)
+			continue;
+
+		session->policy = NULL;
+		assign_policy_plugin(session);
+	}
+}
+
+static gint compare_priority(gconstpointer a, gconstpointer b)
+{
+	const struct connman_session_policy *policy1 = a;
+	const struct connman_session_policy *policy2 = b;
+
+	return policy2->priority - policy1->priority;
+}
+
+
+int connman_session_policy_register(struct connman_session_policy *policy)
+{
+	DBG("name %s", policy->name);
+
+	if (policy->create == NULL || policy->destroy == NULL)
+		return -EINVAL;
+
+	policy_list = g_slist_insert_sorted(policy_list, policy,
+						compare_priority);
+
+	probe_policy(policy);
+
+	return 0;
+}
+
+void connman_session_policy_unregister(struct connman_session_policy *policy)
+{
+	DBG("name %s", policy->name);
+
+	policy_list = g_slist_remove(policy_list, policy);
+
+	remove_policy(policy);
+}
+
+int connman_session_set_default_config(struct connman_session_config *config)
+{
+	config->priority = FALSE;
+	config->roaming_policy = CONNMAN_SESSION_ROAMING_POLICY_DEFAULT;
+	config->type = CONNMAN_SESSION_TYPE_ANY;
+	config->ecall = FALSE;
+
+	g_slist_free(config->allowed_bearers);
+	config->allowed_bearers = g_slist_prepend(NULL,
+				GINT_TO_POINTER(CONNMAN_SERVICE_TYPE_UNKNOWN));
+	if (config->allowed_bearers == NULL)
+		return -ENOMEM;
+
+	return 0;
+}
+
+struct connman_session_config *connman_session_create_default_config(void)
+{
+	struct connman_session_config *config;
+
+	config = g_try_new0(struct connman_session_config, 1);
+	if (config == NULL)
+		return NULL;
+
+	if (connman_session_set_default_config(config) < 0) {
+		g_free(config);
+		return NULL;
+	}
+
+	return config;
+}
+
+static enum connman_session_type apply_policy_on_type(
+			enum connman_session_type policy,
+			enum connman_session_type type)
+{
+	if (type == CONNMAN_SESSION_TYPE_UNKNOWN)
+		return CONNMAN_SESSION_TYPE_UNKNOWN;
+
+	if (policy == CONNMAN_SESSION_TYPE_ANY)
+		return type;
+
+	if (policy == CONNMAN_SESSION_TYPE_LOCAL)
+		return CONNMAN_SESSION_TYPE_LOCAL;
+
+	return CONNMAN_SESSION_TYPE_INTERNET;
+}
+
+int connman_session_parse_bearers(const char *token, GSList **list)
+{
+	enum connman_service_type bearer;
+	int err;
+
+	if (g_strcmp0(token, "") == 0)
+		return 0;
+
+	err = bearer2service(token, &bearer);
+	if (err < 0)
+		return err;
+
+	*list = g_slist_append(*list, GINT_TO_POINTER(bearer));
+
+	return 0;
+}
+
+static int parse_bearers(DBusMessageIter *iter, GSList **list)
 {
-	struct bearer_info *info;
 	DBusMessageIter array;
-	GSList *list = NULL;
+	int type, err;
 
 	dbus_message_iter_recurse(iter, &array);
 
-	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_STRING) {
-		char *bearer = NULL;
+	*list = NULL;
 
-		dbus_message_iter_get_basic(&array, &bearer);
+	while ((type = dbus_message_iter_get_arg_type(&array)) !=
+			DBUS_TYPE_INVALID) {
+		char *bearer_name = NULL;
 
-		info = g_try_new0(struct bearer_info, 1);
-		if (info == NULL) {
-			g_slist_foreach(list, cleanup_bearer_info, NULL);
-			g_slist_free(list);
-
-			return NULL;
+		if (type != DBUS_TYPE_STRING) {
+			g_slist_free(*list);
+			*list = NULL;
+			return -EINVAL;
 		}
 
-		info->name = g_strdup(bearer);
-		info->service_type = bearer2service(info->name);
+		dbus_message_iter_get_basic(&array, &bearer_name);
 
-		if (info->service_type == CONNMAN_SERVICE_TYPE_UNKNOWN &&
-				g_strcmp0(info->name, "*") == 0) {
-			info->match_all = TRUE;
-		} else {
-			info->match_all = FALSE;
+		err = connman_session_parse_bearers(bearer_name, list);
+		if (err < 0) {
+			g_slist_free(*list);
+			*list = NULL;
+			return err;
 		}
 
-		list = g_slist_append(list, info);
-
 		dbus_message_iter_next(&array);
 	}
 
-	return list;
+	return 0;
 }
 
-static GSList *session_allowed_bearers_any(void)
+static int filter_bearer(GSList *policy_bearers,
+				enum connman_service_type bearer,
+				GSList **list)
 {
-	struct bearer_info *info;
-	GSList *list = NULL;
+	enum connman_service_type policy;
+	GSList *it;
 
-	info = g_try_new0(struct bearer_info, 1);
-	if (info == NULL) {
-		g_slist_free(list);
+	if (policy_bearers == NULL)
+		goto clone;
 
-		return NULL;
+	for (it = policy_bearers; it != NULL; it = it->next) {
+		policy = GPOINTER_TO_INT(it->data);
+
+		if (bearer == CONNMAN_SERVICE_TYPE_UNKNOWN) {
+			bearer = policy;
+			goto clone;
+		}
+
+		if (policy != CONNMAN_SERVICE_TYPE_UNKNOWN && policy != bearer)
+			continue;
+
+		goto clone;
 	}
 
-	info->name = g_strdup("");
-	info->match_all = TRUE;
-	info->service_type = CONNMAN_SERVICE_TYPE_UNKNOWN;
+	*list = NULL;
 
-	list = g_slist_append(list, info);
+	return 0;
 
-	return list;
+clone:
+	*list = g_slist_append(*list, GINT_TO_POINTER(bearer));
+
+	return 0;
+}
+
+static int apply_policy_on_bearers(GSList *policy_bearers, GSList *bearers,
+				GSList **list)
+{
+	enum connman_service_type bearer;
+	GSList *it;
+	int err;
+
+	*list = NULL;
+
+	for (it = bearers; it != NULL; it = it->next) {
+		bearer = GPOINTER_TO_INT(it->data);
+
+		err = filter_bearer(policy_bearers, bearer, list);
+		if (err < 0)
+			return err;
+	}
+
+	return 0;
+}
+
+const char *connman_session_get_owner(struct connman_session *session)
+{
+	return session->owner;
 }
 
 static void append_allowed_bearers(DBusMessageIter *iter, void *user_data)
@@ -346,12 +570,16 @@ static void append_allowed_bearers(DBusMessageIter *iter, void *user_data)
 	struct session_info *info = user_data;
 	GSList *list;
 
-	for (list = info->allowed_bearers;
+	for (list = info->config.allowed_bearers;
 			list != NULL; list = list->next) {
-		struct bearer_info *bearer_info = list->data;
+		enum connman_service_type bearer = GPOINTER_TO_INT(list->data);
+		const char *name = __connman_service_type2string(bearer);
+
+		if (name == NULL)
+			name = "*";
 
 		dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING,
-						&bearer_info->name);
+						&name);
 	}
 }
 
@@ -451,30 +679,23 @@ static void append_notify(DBusMessageIter *dict,
 		info_last->entry = info->entry;
 	}
 
-	if (session->append_all == TRUE || info->type != info_last->type) {
-		const char *type = type2string(info->type);
+	if (session->append_all == TRUE ||
+			info->config.type != info_last->config.type) {
+		const char *type = type2string(info->config.type);
 
 		connman_dbus_dict_append_basic(dict, "ConnectionType",
 						DBUS_TYPE_STRING,
 						&type);
-		info_last->type = info->type;
+		info_last->config.type = info->config.type;
 	}
 
 	if (session->append_all == TRUE ||
-			info->allowed_bearers != info_last->allowed_bearers) {
+			info->config.allowed_bearers != info_last->config.allowed_bearers) {
 		connman_dbus_dict_append_array(dict, "AllowedBearers",
 						DBUS_TYPE_STRING,
 						append_allowed_bearers,
 						info);
-		info_last->allowed_bearers = info->allowed_bearers;
-	}
-
-	if (session->append_all == TRUE ||
-			info->avoid_handover != info_last->avoid_handover) {
-		connman_dbus_dict_append_basic(dict, "AvoidHandover",
-						DBUS_TYPE_BOOLEAN,
-						&info->avoid_handover);
-		info_last->avoid_handover = info->avoid_handover;
+		info_last->config.allowed_bearers = info->config.allowed_bearers;
 	}
 
 	session->append_all = FALSE;
@@ -484,6 +705,8 @@ static connman_bool_t is_type_matching_state(enum connman_session_state *state,
 						enum connman_session_type type)
 {
 	switch (type) {
+	case CONNMAN_SESSION_TYPE_UNKNOWN:
+		return FALSE;
 	case CONNMAN_SESSION_TYPE_ANY:
 		return TRUE;
 	case CONNMAN_SESSION_TYPE_LOCAL:
@@ -517,9 +740,8 @@ static connman_bool_t compute_notifiable_changes(struct connman_session *session
 			info->state >= CONNMAN_SESSION_STATE_CONNECTED)
 		return TRUE;
 
-	if (info->allowed_bearers != info_last->allowed_bearers ||
-			info->avoid_handover != info_last->avoid_handover ||
-			info->type != info_last->type)
+	if (info->config.allowed_bearers != info_last->config.allowed_bearers ||
+			info->config.type != info_last->config.type)
 		return TRUE;
 
 	return FALSE;
@@ -579,16 +801,16 @@ static connman_bool_t service_type_match(struct connman_session *session,
 	struct session_info *info = session->info;
 	GSList *list;
 
-	for (list = info->allowed_bearers;
+	for (list = info->config.allowed_bearers;
 			list != NULL; list = list->next) {
-		struct bearer_info *bearer_info = list->data;
+		enum connman_service_type bearer = GPOINTER_TO_INT(list->data);
 		enum connman_service_type service_type;
 
-		if (bearer_info->match_all == TRUE)
+		if (bearer == CONNMAN_SERVICE_TYPE_UNKNOWN)
 			return TRUE;
 
 		service_type = connman_service_get_type(service);
-		if (bearer_info->service_type == service_type)
+		if (bearer == service_type)
 			return TRUE;
 	}
 
@@ -613,7 +835,7 @@ static int service_type_weight(enum connman_service_type type)
 	 *
 	 * 1. Ethernet
 	 * 2. Bluetooth
-	 * 3. WiFi/WiMAX
+	 * 3. WiFi
 	 * 4. Cellular
 	 */
 
@@ -623,7 +845,6 @@ static int service_type_weight(enum connman_service_type type)
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 		return 3;
 	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
 		return 2;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		return 1;
@@ -650,11 +871,11 @@ static gint sort_allowed_bearers(struct connman_service *service_a,
 	type_a = connman_service_get_type(service_a);
 	type_b = connman_service_get_type(service_b);
 
-	for (list = info->allowed_bearers;
+	for (list = info->config.allowed_bearers;
 			list != NULL; list = list->next) {
-		struct bearer_info *bearer_info = list->data;
+		enum connman_service_type bearer = GPOINTER_TO_INT(list->data);
 
-		if (bearer_info->match_all == TRUE) {
+		if (bearer == CONNMAN_SERVICE_TYPE_UNKNOWN) {
 			if (type_a != type_b) {
 				weight_a = service_type_weight(type_a);
 				weight_b = service_type_weight(type_b);
@@ -669,20 +890,14 @@ static gint sort_allowed_bearers(struct connman_service *service_a,
 			}
 		}
 
-		if (type_a == bearer_info->service_type &&
-				type_b == bearer_info->service_type) {
+		if (type_a == bearer && type_b == bearer)
 			return 0;
-		}
 
-		if (type_a == bearer_info->service_type &&
-				type_b != bearer_info->service_type) {
+		if (type_a == bearer &&	type_b != bearer)
 			return -1;
-		}
 
-		if (type_a != bearer_info->service_type &&
-				type_b == bearer_info->service_type) {
+		if (type_a != bearer &&	type_b == bearer)
 			return 1;
-		}
 	}
 
 	return 0;
@@ -698,33 +913,6 @@ static gint sort_services(gconstpointer a, gconstpointer b, gpointer user_data)
 				session);
 }
 
-static void cleanup_session(gpointer user_data)
-{
-	struct connman_session *session = user_data;
-	struct session_info *info = session->info;
-
-	DBG("remove %s", session->session_path);
-
-	g_hash_table_destroy(session->service_hash);
-	g_sequence_free(session->service_list);
-
-	if (info->entry != NULL &&
-			info->entry->reason == CONNMAN_SESSION_REASON_CONNECT) {
-		__connman_service_disconnect(info->entry->service);
-	}
-
-	g_slist_foreach(info->allowed_bearers, cleanup_bearer_info, NULL);
-	g_slist_free(info->allowed_bearers);
-
-	g_free(session->owner);
-	g_free(session->session_path);
-	g_free(session->notify_path);
-	g_free(session->info);
-	g_free(session->info_last);
-
-	g_free(session);
-}
-
 static enum connman_session_state service_to_session_state(enum connman_service_state state)
 {
 	switch (state) {
@@ -950,7 +1138,7 @@ static void select_connected_service(struct session_info *info,
 	enum connman_session_state state;
 
 	state = service_to_session_state(entry->state);
-	if (is_type_matching_state(&state, info->type) == FALSE)
+	if (is_type_matching_state(&state, info->config.type) == FALSE)
 		return;
 
 	info->state = state;
@@ -1118,7 +1306,7 @@ static void session_changed(struct connman_session *session,
 
 		state = service_to_session_state(info->entry->state);
 
-		if (is_type_matching_state(&state, info->type) == TRUE)
+		if (is_type_matching_state(&state, info->config.type) == TRUE)
 			info->state = state;
 	}
 
@@ -1127,7 +1315,7 @@ static void session_changed(struct connman_session *session,
 		DBG("ignore session changed event");
 		return;
 	case CONNMAN_SESSION_TRIGGER_SETTING:
-		if (info->allowed_bearers != info_last->allowed_bearers) {
+		if (info->config.allowed_bearers != info_last->config.allowed_bearers) {
 
 			service_hash_last = session->service_hash;
 			service_list_last = session->service_list;
@@ -1155,10 +1343,10 @@ static void session_changed(struct connman_session *session,
 			g_sequence_free(service_list_last);
 		}
 
-		if (info->type != info_last->type) {
+		if (info->config.type != info_last->config.type) {
 			if (info->state >= CONNMAN_SESSION_STATE_CONNECTED &&
 					is_type_matching_state(&info->state,
-							info->type) == FALSE)
+							info->config.type) == FALSE)
 				deselect_and_disconnect(session);
 		}
 
@@ -1214,6 +1402,46 @@ static void session_changed(struct connman_session *session,
 	session_notify(session);
 }
 
+int connman_session_config_update(struct connman_session *session)
+{
+	struct session_info *info = session->info;
+	GSList *allowed_bearers;
+	int err;
+
+	DBG("session %p", session);
+
+	/*
+	 * We update all configuration even though only one entry
+	 * might have changed. We can still optimize this later.
+	 */
+
+	err = apply_policy_on_bearers(
+		session->policy_config->allowed_bearers,
+		session->user_allowed_bearers,
+		&allowed_bearers);
+	if (err < 0)
+		return err;
+
+	g_slist_free(info->config.allowed_bearers);
+	info->config.allowed_bearers = allowed_bearers;
+
+	info->config.type = apply_policy_on_type(
+				session->policy_config->type,
+				info->config.type);
+
+	info->config.roaming_policy = session->policy_config->roaming_policy;
+
+	info->config.ecall = session->policy_config->ecall;
+	if (info->config.ecall == TRUE)
+		ecall_session = session;
+
+	info->config.priority = session->policy_config->priority;
+
+	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
+
+	return 0;
+}
+
 static DBusMessage *connect_session(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -1261,40 +1489,40 @@ static DBusMessage *change_session(DBusConnection *conn,
 	const char *name;
 	const char *val;
 	GSList *allowed_bearers;
+	int err;
 
 	DBG("session %p", session);
 	if (dbus_message_iter_init(msg, &iter) == FALSE)
 		return __connman_error_invalid_arguments(msg);
 
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
+		return __connman_error_invalid_arguments(msg);
+
 	dbus_message_iter_get_basic(&iter, &name);
 	dbus_message_iter_next(&iter);
+
+	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_VARIANT)
+		return __connman_error_invalid_arguments(msg);
+
 	dbus_message_iter_recurse(&iter, &value);
 
 	switch (dbus_message_iter_get_arg_type(&value)) {
 	case DBUS_TYPE_ARRAY:
 		if (g_str_equal(name, "AllowedBearers") == TRUE) {
-			allowed_bearers = session_parse_allowed_bearers(&value);
+			err = parse_bearers(&value, &allowed_bearers);
+			if (err < 0)
+				return __connman_error_failed(msg, err);
 
-			g_slist_foreach(info->allowed_bearers,
-					cleanup_bearer_info, NULL);
-			g_slist_free(info->allowed_bearers);
+			g_slist_free(info->config.allowed_bearers);
+			session->user_allowed_bearers = allowed_bearers;
 
-			if (allowed_bearers == NULL) {
-				allowed_bearers = session_allowed_bearers_any();
+			err = apply_policy_on_bearers(
+					session->policy_config->allowed_bearers,
+					session->user_allowed_bearers,
+					&info->config.allowed_bearers);
 
-				if (allowed_bearers == NULL)
-					return __connman_error_failed(msg, ENOMEM);
-			}
-
-			info->allowed_bearers = allowed_bearers;
-		} else {
-			goto err;
-		}
-		break;
-	case DBUS_TYPE_BOOLEAN:
-		if (g_str_equal(name, "AvoidHandover") == TRUE) {
-			dbus_message_iter_get_basic(&value,
-					&info->avoid_handover);
+			if (err < 0)
+				return __connman_error_failed(msg, err);
 		} else {
 			goto err;
 		}
@@ -1302,7 +1530,9 @@ static DBusMessage *change_session(DBusConnection *conn,
 	case DBUS_TYPE_STRING:
 		if (g_str_equal(name, "ConnectionType") == TRUE) {
 			dbus_message_iter_get_basic(&value, &val);
-			info->type = string2type(val);
+			info->config.type = apply_policy_on_type(
+				session->policy_config->type,
+				connman_session_parse_connection_type(val));
 		} else {
 			goto err;
 		}
@@ -1396,20 +1626,99 @@ static const GDBusMethodTable session_methods[] = {
 	{ },
 };
 
+static int session_create_cb(struct connman_session *session,
+				struct connman_session_config *config,
+				void *user_data, int err)
+{
+	DBusMessage *reply;
+	struct user_config *user_config = user_data;
+	struct session_info *info, *info_last;
+
+	DBG("session %p config %p", session, config);
+
+	if (err != 0)
+		goto out;
+
+	session->policy_config = config;
+
+	info = session->info;
+	info_last = session->info_last;
+
+	if (session->policy_config->ecall == TRUE)
+		ecall_session = session;
+
+	info->state = CONNMAN_SESSION_STATE_DISCONNECTED;
+	info->config.type = apply_policy_on_type(
+				session->policy_config->type,
+				user_config->type);
+	info->config.priority = session->policy_config->priority;
+	info->config.roaming_policy = session->policy_config->roaming_policy;
+	info->entry = NULL;
+
+	session->user_allowed_bearers = user_config->allowed_bearers;
+	user_config->allowed_bearers = NULL;
+
+	err = apply_policy_on_bearers(
+			session->policy_config->allowed_bearers,
+			session->user_allowed_bearers,
+			&info->config.allowed_bearers);
+	if (err < 0)
+		goto out;
+
+	g_hash_table_replace(session_hash, session->session_path, session);
+
+	DBG("add %s", session->session_path);
+
+	if (g_dbus_register_interface(connection, session->session_path,
+					CONNMAN_SESSION_INTERFACE,
+					session_methods, NULL,
+					NULL, session, NULL) == FALSE) {
+		connman_error("Failed to register %s", session->session_path);
+		g_hash_table_remove(session_hash, session->session_path);
+		err = -EINVAL;
+		goto out;
+	}
+
+	reply = g_dbus_create_reply(user_config->pending,
+				DBUS_TYPE_OBJECT_PATH, &session->session_path,
+				DBUS_TYPE_INVALID);
+	g_dbus_send_message(connection, reply);
+	user_config->pending = NULL;
+
+	populate_service_list(session);
+
+	info_last->state = info->state;
+	info_last->config.priority = info->config.priority;
+	info_last->config.roaming_policy = info->config.roaming_policy;
+	info_last->entry = info->entry;
+	info_last->config.allowed_bearers = info->config.allowed_bearers;
+
+	session->append_all = TRUE;
+
+	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
+
+out:
+	if (err < 0) {
+		reply = __connman_error_failed(user_config->pending, -err);
+		g_dbus_send_message(connection, reply);
+
+		free_session(session);
+	}
+
+	cleanup_user_config(user_config);
+
+	return err;
+}
+
 int __connman_session_create(DBusMessage *msg)
 {
 	const char *owner, *notify_path;
 	char *session_path = NULL;
 	DBusMessageIter iter, array;
 	struct connman_session *session = NULL;
-	struct session_info *info, *info_last;
-
-	enum connman_session_type type = CONNMAN_SESSION_TYPE_ANY;
-	connman_bool_t priority, avoid_handover = FALSE;
-	connman_bool_t ecall_app;
-	enum connman_session_roaming_policy roaming_policy;
-	char *roaming_policy_str;
-	GSList *allowed_bearers = NULL;
+	struct user_config *user_config = NULL;
+	connman_bool_t user_allowed_bearers = FALSE;
+	connman_bool_t user_connection_type = FALSE;
 	int err;
 
 	owner = dbus_message_get_sender(msg);
@@ -1425,6 +1734,14 @@ int __connman_session_create(DBusMessage *msg)
 		goto err;
 	}
 
+	user_config = g_try_new0(struct user_config, 1);
+	if (user_config == NULL) {
+		err = -ENOMEM;
+		goto err;
+	}
+
+	user_config->pending = dbus_message_ref(msg);
+
 	dbus_message_iter_init(msg, &iter);
 	dbus_message_iter_recurse(&iter, &array);
 
@@ -1441,31 +1758,52 @@ int __connman_session_create(DBusMessage *msg)
 		switch (dbus_message_iter_get_arg_type(&value)) {
 		case DBUS_TYPE_ARRAY:
 			if (g_str_equal(key, "AllowedBearers") == TRUE) {
-				allowed_bearers =
-					session_parse_allowed_bearers(&value);
+				err = parse_bearers(&value,
+						&user_config->allowed_bearers);
+				if (err < 0)
+					goto err;
+
+				user_allowed_bearers = TRUE;
 			} else {
-				return -EINVAL;
-			}
-			break;
-		case DBUS_TYPE_BOOLEAN:
-			if (g_str_equal(key, "AvoidHandover") == TRUE) {
-				dbus_message_iter_get_basic(&value,
-							&avoid_handover);
-			} else {
-				return -EINVAL;
+				err = -EINVAL;
+				goto err;
 			}
 			break;
 		case DBUS_TYPE_STRING:
 			if (g_str_equal(key, "ConnectionType") == TRUE) {
 				dbus_message_iter_get_basic(&value, &val);
-				type = string2type(val);
+				user_config->type =
+					connman_session_parse_connection_type(val);
+
+				user_connection_type = TRUE;
 			} else {
-				return -EINVAL;
+				err = -EINVAL;
+				goto err;
 			}
 		}
 		dbus_message_iter_next(&array);
 	}
 
+	/*
+	 * If the user hasn't provided a configuration, we set
+	 * the default configuration.
+	 *
+	 * For AllowedBearers this is '*', ...
+	 */
+	if (user_allowed_bearers == FALSE) {
+		user_config->allowed_bearers =
+			g_slist_append(NULL,
+				GINT_TO_POINTER(CONNMAN_SERVICE_TYPE_UNKNOWN));
+		if (user_config->allowed_bearers == NULL) {
+			err = -ENOMEM;
+			goto err;
+		}
+	}
+
+	/* ... and for ConnectionType it is 'any'. */
+	if (user_connection_type == FALSE)
+		user_config->type = CONNMAN_SESSION_TYPE_ANY;
+
 	dbus_message_iter_next(&iter);
 	dbus_message_iter_get_basic(&iter, &notify_path);
 
@@ -1482,6 +1820,7 @@ int __connman_session_create(DBusMessage *msg)
 
 	session = g_hash_table_lookup(session_hash, session_path);
 	if (session != NULL) {
+		g_free(session_path);
 		session = NULL;
 		err = -EEXIST;
 		goto err;
@@ -1489,10 +1828,13 @@ int __connman_session_create(DBusMessage *msg)
 
 	session = g_try_new0(struct connman_session, 1);
 	if (session == NULL) {
+		g_free(session_path);
 		err = -ENOMEM;
 		goto err;
 	}
 
+	session->session_path = session_path;
+
 	session->info = g_try_new0(struct session_info, 1);
 	if (session->info == NULL) {
 		err = -ENOMEM;
@@ -1505,97 +1847,38 @@ int __connman_session_create(DBusMessage *msg)
 		goto err;
 	}
 
-	info = session->info;
-	info_last = session->info_last;
-
-	config_get_bool(owner, "Priority", &priority);
-	config_get_bool(owner, "EmergencyCall", &ecall_app);
-	config_get_string(owner, "RoamingPolicy", &roaming_policy_str);
-	roaming_policy = string2roamingpolicy(roaming_policy_str);
-
 	session->owner = g_strdup(owner);
-	session->session_path = session_path;
 	session->notify_path = g_strdup(notify_path);
 	session->notify_watch =
 		g_dbus_add_disconnect_watch(connection, session->owner,
 					owner_disconnect, session, NULL);
-	if (ecall_app == TRUE)
-		ecall_session = session;
 
-	info->state = CONNMAN_SESSION_STATE_DISCONNECTED;
-	info->type = type;
-	info->priority = priority;
-	info->avoid_handover = avoid_handover;
-	info->roaming_policy = roaming_policy;
-	info->entry = NULL;
-
-	if (allowed_bearers == NULL) {
-		info->allowed_bearers =
-				session_allowed_bearers_any();
-
-		if (info->allowed_bearers == NULL) {
-			err = -ENOMEM;
-			goto err;
-		}
-	} else {
-		info->allowed_bearers = allowed_bearers;
-	}
-
-	g_hash_table_replace(session_hash, session->session_path, session);
-
-	DBG("add %s", session->session_path);
-
-	if (g_dbus_register_interface(connection, session->session_path,
-					CONNMAN_SESSION_INTERFACE,
-					session_methods, NULL,
-					NULL, session, NULL) == FALSE) {
-		connman_error("Failed to register %s", session->session_path);
-		g_hash_table_remove(session_hash, session->session_path);
-		session = NULL;
-
-		err = -EINVAL;
+	err = assign_policy_plugin(session);
+	if (err < 0)
 		goto err;
-	}
 
-	g_dbus_send_reply(connection, msg,
-				DBUS_TYPE_OBJECT_PATH, &session->session_path,
-				DBUS_TYPE_INVALID);
+	err = create_policy_config(session, session_create_cb, user_config);
+	if (err < 0 && err != -EINPROGRESS)
+		return err;
 
-
-	populate_service_list(session);
-
-	info_last->state = info->state;
-	info_last->priority = info->priority;
-	info_last->avoid_handover = info->avoid_handover;
-	info_last->roaming_policy = info->roaming_policy;
-	info_last->entry = info->entry;
-	info_last->allowed_bearers = info->allowed_bearers;
-
-	session->append_all = TRUE;
-
-	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
-
-	return 0;
+	return -EINPROGRESS;
 
 err:
 	connman_error("Failed to create session");
 
-	if (session != NULL) {
-		if (session->info_last != NULL)
-			g_free(session->info_last);
-		if (session->info != NULL)
-			g_free(session->info);
-		g_free(session);
-	}
-
-	g_free(session_path);
-
-	g_slist_foreach(allowed_bearers, cleanup_bearer_info, NULL);
-	g_slist_free(allowed_bearers);
+	free_session(session);
 
+	cleanup_user_config(user_config);
 	return err;
 }
 
+void connman_session_destroy(struct connman_session *session)
+{
+	DBG("session %p", session);
+
+	session_disconnect(session);
+}
+
 int __connman_session_destroy(DBusMessage *msg)
 {
 	const char *owner, *session_path;
@@ -1617,7 +1900,7 @@ int __connman_session_destroy(DBusMessage *msg)
 	if (g_strcmp0(owner, session->owner) != 0)
 		return -EACCES;
 
-	session_disconnect(session);
+	connman_session_destroy(session);
 
 	return 0;
 }
@@ -1774,98 +2057,6 @@ static struct connman_notifier session_notifier = {
 	.ipconfig_changed	= ipconfig_changed,
 };
 
-static struct connman_session *session_lookup_by_id(const char *id)
-{
-	struct connman_session *session;
-	GHashTableIter iter;
-	gpointer key, value;
-
-	DBG("id %s", id);
-
-	g_hash_table_iter_init(&iter, session_hash);
-
-	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
-		session = value;
-
-		if (g_strcmp0(session->owner, id) == FALSE)
-			continue;
-
-		return session;
-	}
-
-	DBG("No session found by id %s", id);
-
-	return NULL;
-}
-
-int connman_session_update_bool(const char *id, const char *key,
-					connman_bool_t val)
-{
-	struct connman_session *session;
-	struct session_info *info;
-
-	session = session_lookup_by_id(id);
-	if (session == NULL)
-		return -EINVAL;
-
-	info = session->info;
-	if (info == NULL)
-		return 0;
-
-	DBG("%s %d", key, val);
-
-	return -EINVAL;
-}
-
-int connman_session_update_string(const char *id, const char *key,
-					const char *val)
-{
-	struct connman_session *session;
-	struct session_info *info;
-
-	session = session_lookup_by_id(id);
-	if (session == NULL)
-		return -EINVAL;
-
-	info = session->info;
-	if (info == NULL)
-		return 0;
-
-	DBG("%s %s", key, val);
-
-	return -EINVAL;
-}
-
-int connman_session_config_register(struct connman_session_config *config)
-{
-	DBG("name %s", config->name);
-
-	if (session_config != NULL) {
-		connman_warn("A session configuration plugin '%s' is "
-				"already registerd. Skipping registration "
-				"of plugin '%s'",
-				session_config->name, config->name);
-		return -EALREADY;
-	}
-
-	session_config = config;
-
-	return 0;
-}
-
-void connman_session_config_unregister(struct connman_session_config *config)
-{
-	DBG("name %s", config->name);
-
-	if (config != session_config) {
-		connman_warn("Trying to unregister session configuration "
-				"plugin '%s'", config->name);
-		return;
-	}
-
-	session_config = NULL;
-}
-
 int __connman_session_init(void)
 {
 	int err;
